//  gmodconfig - a Gnome based kernel module configuration tool
//  http://gmodconfig.sourceforge.net/
//  Copyright (C) 2003 Cyril Bortolato
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  modgui.cpp: build GUI of a module
//--------------------------------------------------------------------------//

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <iterator>
#include <sstream>
#include <iostream>
#include <algorithm>
#include <vector>
#include "installmgr.h"
#include "modgui.h"
#include "moduletypes.h"
#include "support.h"
#include "xsatypes.h"
#include "xsagui.h"

using namespace gmodcfg;
using namespace std;

// Notebook widget's tab order
enum { PARAMETERS_TAB = 0, SETUP_TAB, UPDATES_TAB, ABOUT_TAB };

//--------------------------------------------------------------------------//
//  Create a left justified label in a table
//--------------------------------------------------------------------------//
static GtkWidget*
makeLabel( GtkTable* table, const string& str, int row, int col )
{
    GtkWidget* label = gtk_label_new( str.c_str() );
    gtk_label_set_justify( GTK_LABEL(label), GTK_JUSTIFY_LEFT );
    gtk_label_set_line_wrap( GTK_LABEL(label), TRUE );
    gtk_misc_set_alignment( GTK_MISC(label), 0, 0.5 );
    gtk_widget_show( label );
    gtk_table_attach( table, label, col, col+1, row, row+1,
                      (GtkAttachOptions)(GTK_FILL),
                      (GtkAttachOptions)(GTK_FILL), 1 , 1 );
    return label;
}

//--------------------------------------------------------------------------//
//  Create a window with automatic scrollbars
//--------------------------------------------------------------------------//
static GtkWidget*
makeScrolledWindow()
{
    GtkWidget* scroll = gtk_scrolled_window_new(NULL, NULL);
    gtk_scrolled_window_set_policy( GTK_SCROLLED_WINDOW(scroll),
                                    GTK_POLICY_AUTOMATIC,
                                    GTK_POLICY_AUTOMATIC );
    gtk_widget_show( scroll );
    return scroll;
}

//--------------------------------------------------------------------------//
//  Create a table inside a scrolled window
//--------------------------------------------------------------------------//
static GtkTable*
makeTable( GtkWidget* scroll, int numRows, int numCols )
{
    GtkWidget* table = gtk_table_new( numRows, numCols, FALSE );
    gtk_scrolled_window_add_with_viewport( GTK_SCROLLED_WINDOW(scroll), table );
    gtk_widget_show( table );
    return GTK_TABLE( table );
}

//--------------------------------------------------------------------------//
//  Add a widget to a table cell, and align it on the left
//--------------------------------------------------------------------------//
static void
addToTable( GtkTable* table, GtkWidget* widget, int row, int col,
            int xpadding = 0, int ypadding = 0 )
{
    GtkWidget* align = gtk_alignment_new( 0, 0.5, 0, 0 );
    gtk_container_add( GTK_CONTAINER(align), widget );
    gtk_widget_show( align );
    gtk_table_attach( table, align, col, col+1, row, row+1,
                      (GtkAttachOptions)(GTK_FILL),
                      (GtkAttachOptions)(GTK_FILL), xpadding, ypadding );
}

//--------------------------------------------------------------------------//
//  Create an hyperlink
//--------------------------------------------------------------------------//
static GtkWidget*
makeHyperlink( GtkTable* table, const string& url, const string& msg,
               int row, int col )
{
    GtkWidget* href = gnome_href_new( url.c_str(), msg.c_str() );
    gtk_widget_show( href );
    addToTable( table, href, row, col );
    return href;
}

//--------------------------------------------------------------------------//
//  Replace a tab's content by a new widget 
//--------------------------------------------------------------------------//
static void
replaceTab( GtkNotebook* notebook, gint tabNum, GtkWidget* newTab )
{
    GtkWidget* oldpage = gtk_notebook_get_nth_page( notebook, tabNum );
    GtkWidget* label = gtk_notebook_get_tab_label( notebook, oldpage );
    gtk_notebook_remove_page( notebook, tabNum );
    gtk_notebook_insert_page( notebook, newTab, label, tabNum );
}

//--------------------------------------------------------------------------//
//  Build a checkbox for a bool parameter
//--------------------------------------------------------------------------//
static GtkWidget*
buildBoolParam( const string& paramVal )
{
    GtkWidget* checkbox = gtk_check_button_new();
    if ( paramVal == "1" )
        gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON(checkbox), TRUE );
    gtk_widget_show( checkbox );
    return checkbox;
}

//--------------------------------------------------------------------------//
//  Build an entry or a combo
//--------------------------------------------------------------------------//
static GtkWidget*
buildEntryParam( const string& paramVal, const string& values,
                 int verbosity = 0 )
{
    GtkWidget* entry = NULL;
    GtkWidget* parent = NULL;

    if ( values.empty() )
        parent = entry = gtk_entry_new();
    else
    {
        parent = gtk_combo_new();
        entry = GTK_COMBO(parent)->entry;
        gtk_combo_set_value_in_list( GTK_COMBO(parent), TRUE, TRUE );

        if ( verbosity >= 3 )
            cout << "values = \"" << values << "\"" << endl;

        // Populate list of values
        istringstream istr( values );
        istream_iterator<string> vit( istr ), vitend;
        vector<string> valvect;
        copy( vit, vitend, back_inserter(valvect) );

        GList* vallist = NULL;

        // If no default value, create an empty one
        if ( paramVal.empty() )
            vallist = g_list_append( vallist, (gpointer) "" );

        for( unsigned int i=0; i<valvect.size(); ++i )
        {
            if ( verbosity >= 3 )
                cout << "list value = " << valvect[i] << endl;
            vallist = g_list_append( vallist, (gpointer) valvect[i].c_str() );
        }

        gtk_combo_set_popdown_strings( GTK_COMBO(parent), vallist );
        g_list_free( vallist );
    }

    if ( !paramVal.empty() )
        gtk_entry_set_text( GTK_ENTRY(entry), paramVal.c_str() );

    gtk_widget_set_size_request( parent, 50, -1 );
    gtk_widget_show( parent );
    return parent;
}

//--------------------------------------------------------------------------//
//  Build a combo with all values in the range (array parameter)
//--------------------------------------------------------------------------//
static GtkWidget*
buildArrayParam( const string& paramVal, int minVal, int maxVal )
{
    ostringstream oss;
    for( int i=minVal; i<=maxVal; ++i )
        oss << i << " ";

    return buildEntryParam( paramVal, oss.str() );
}

//--------------------------------------------------------------------------//
//  Build the Parameters tab in the notebook widget
//--------------------------------------------------------------------------//
void
ModuleGuiBuilder::buildParametersTab( const Module* mod,
                                      const StringStringMap_t& optMap )
{
    GtkWidget* scroll = makeScrolledWindow();
    GtkTable* table = makeTable( scroll, mod->GetNumParams(), 2 );

    const string& lang = mod->GetLanguage();

    ParamVector::const_iterator it = mod->BeginParams();
    ParamVector::const_iterator itend = mod->EndParams();
    for( int row=0; it != itend; ++row, ++it )
    {
        const Param* p = *it;
        string desc = p->GetDescription( lang );
        if ( desc.empty() )
            desc = p->GetDescription( "en" ); // use English as fallback

        // Get value already assigned to this parameter, or the default
        string paramVal = p->GetDefault();
        StringStringMap_t::const_iterator mit = optMap.find( p->GetName() );
        if ( mit != optMap.end() )
            paramVal = (*mit).second;

        GtkWidget* paramWidget = 0;

        switch( p->GetType() )
        {
            case PARAM_BOOL:
            {
                paramWidget = buildBoolParam( paramVal );
                break;
            }

            case PARAM_ARRAY:
            {
                int minVal = atoi( p->GetMin().c_str() );
                int maxVal = atoi( p->GetMax().c_str() );
                paramWidget = buildArrayParam( paramVal, minVal, maxVal );
                break;
            }

            case PARAM_INT:
            case PARAM_LONG:
            case PARAM_SHORT:
            case PARAM_CHAR:
            case PARAM_STRING:
            {
                paramWidget = buildEntryParam( paramVal, p->GetValues(),
                                               Verbosity );
                break;
            }

            default:
                assert(0); // unknown type
        }

        addToTable( table, paramWidget, row, 0 );
        makeLabel( table, desc.c_str(), row, 1 );

        // Keep track of entry widgets to query their content later
        if ( GTK_IS_COMBO(paramWidget) )
            paramWidget = GTK_COMBO(paramWidget)->entry;
        WidgetMap[ p ] = paramWidget;
    }

    if ( mod->GetNumParams() == 0 )
        makeLabel( table, _("This module has no parameters"), 0, 0 );

    replaceTab( Notebook, PARAMETERS_TAB, scroll );
}

//--------------------------------------------------------------------------//
//  Build the Setup tab in the notebook widget
//--------------------------------------------------------------------------//
void
ModuleGuiBuilder::buildSetupTab( const Module* mod )
{
    GtkWidget* scroll = makeScrolledWindow();
    GtkTable* table = makeTable( scroll, 1, 1 );

    // TODO build setup tab
    makeLabel( table, _("This feature is not implemented yet"), 0, 0 );
    replaceTab( Notebook, SETUP_TAB, scroll );
}

//--------------------------------------------------------------------------//
//  "changed" callback for the versions list treeview in the Updates tab
//--------------------------------------------------------------------------//
static GtkTextBuffer* changesViewerBuffer = 0;
static GtkWidget* installButton = 0;

static void
selectionChangedCallback( GtkTreeSelection* select, gpointer )
{
    GtkTreeView* treeView = gtk_tree_selection_get_tree_view( select );
    GtkTreeModel* model = gtk_tree_view_get_model( treeView );
    GtkTreeIter iter;

    if ( gtk_tree_selection_get_selected( select, &model, &iter ) )
    {
        gchar *changes, *file = 0;
        gtk_tree_model_get( model, &iter,
                            XSA_PRODUCT_CHANGES, &changes,
                            XSA_PRODUCT_FILE, &file, -1 );
        if ( file && strcmp(file, "") != 0 )
            gtk_widget_set_sensitive( installButton, TRUE );
        else
            gtk_widget_set_sensitive( installButton, FALSE );

        if ( changes )
            gtk_text_buffer_set_text( changesViewerBuffer, changes, -1 );
    }
}

//--------------------------------------------------------------------------//
//  "clicked" callback for the install button in the Updates tab
//--------------------------------------------------------------------------//
static void
installButtonClickedCallback( GtkButton*, gpointer data )
{
    XsaTreeViewManager* treeViewMgr = static_cast<XsaTreeViewManager*>(data);
    const gchar* file = treeViewMgr->GetSelectedFile();
    if ( file )
        InstallManager::GetInstallManager()->StartInstall( file );
}

//--------------------------------------------------------------------------//
//  Build the Updates tab in the notebook widget
//--------------------------------------------------------------------------//
void
ModuleGuiBuilder::buildUpdatesTab( const Module* mod )
{
    GtkWidget* scroll = makeScrolledWindow();
    const XsaData* xsaData = mod->GetXsaData();

    if ( !xsaData )
    {
        GtkTable* table = makeTable( scroll, 1, 1 );
        makeLabel( table, _("I'm sorry, new versions informations for this module are not available"), 0, 0 );
        replaceTab( Notebook, UPDATES_TAB, scroll );

        return;
    }

    // Display XSA infos in a table
    GtkTable* table = makeTable( scroll, 5, 1 );

    // Row #0: a label
    GtkWidget* label1 = makeLabel( table,
                          _("The latest versions of this module are:"), 0, 0 );
    gtk_misc_set_padding( GTK_MISC( label1 ), 0, 10 );

    // Row #1: display versions as a list
    GtkWidget* treeView = gtk_tree_view_new();
    addToTable( table, treeView, 1, 0 );
    gtk_widget_show( treeView );

    delete TreeViewMgr;
    TreeViewMgr = new XsaTreeViewManager( GTK_TREE_VIEW(treeView), xsaData );
    TreeViewMgr->InitListStore();

    // Row #2: a label
    GtkWidget* label2 = makeLabel( table, _("Changes:"), 2, 0 );
    gtk_misc_set_padding( GTK_MISC( label2 ), 0, 10 );

    // Row #3: display changes of a version in a textview inside a scrollwindow
    GtkWidget* changesScroll = makeScrolledWindow();
    addToTable( table, changesScroll, 3, 0 );
    GtkWidget* changesViewer = gtk_text_view_new();
    gtk_container_add( GTK_CONTAINER(changesScroll), changesViewer );
    gtk_text_view_set_editable( GTK_TEXT_VIEW(changesViewer), FALSE );
    gtk_text_view_set_wrap_mode( GTK_TEXT_VIEW(changesViewer), GTK_WRAP_WORD );
    gtk_text_view_set_cursor_visible( GTK_TEXT_VIEW(changesViewer), FALSE );
    gtk_widget_set_size_request( changesViewer, 300, -1 );
    gtk_widget_show( changesViewer );
    changesViewerBuffer =
            gtk_text_view_get_buffer( GTK_TEXT_VIEW(changesViewer ) );
  
    // Row #4: add an Install button, initially inactive
    GtkWidget* buttonBox = gtk_vbutton_box_new();
    addToTable( table, buttonBox, 4, 0, 0, 5 );
    gtk_widget_show( buttonBox );
    installButton = gtk_button_new_with_label( _("Install") );
    gtk_container_add( GTK_CONTAINER(buttonBox), installButton );
    gtk_widget_set_sensitive( installButton, FALSE );
    gtk_widget_show( installButton );

    // Activate button when an entry in the list is selected AND has a file
    // info (else we won't know where to fetch the file to install!)
    GtkTreeSelection* select = gtk_tree_view_get_selection(
                                                    GTK_TREE_VIEW(treeView) );
    g_signal_connect( G_OBJECT( select ), "changed",
                      G_CALLBACK( selectionChangedCallback ), NULL );

    // Launch the install manager when button is clicked
    g_signal_connect( G_OBJECT( installButton ), "clicked",
                      G_CALLBACK( installButtonClickedCallback ),
                      TreeViewMgr );

    replaceTab( Notebook, UPDATES_TAB, scroll );
}

//--------------------------------------------------------------------------//
//  Build the About tab in the notebook widget
//--------------------------------------------------------------------------//
void
ModuleGuiBuilder::buildAboutTab( const Module* mod )
{
    GtkWidget* scroll = makeScrolledWindow();
    GtkTable* table = makeTable( scroll, 4, 2 );

    int row = 0;
    makeLabel( table, _("Author:"), row, 0 );
    makeLabel( table, mod->GetAuthors(), row++, 1 );

    makeLabel( table, _("License:"), row, 0 );
    if ( mod->GetLicense() == "GPL" )
    {
        // TRANSLATORS: make URL point to translation of GPL
        makeHyperlink( table, _("http://www.gnu.org/copyleft/gpl.html"),
                       mod->GetLicense(), row++, 1 );
    }
    else
        makeLabel( table, mod->GetLicense(), row++, 1 );

    if ( !mod->GetURL().empty() )
    {
        makeLabel( table, _("Website:"), row, 0 );
        makeHyperlink( table, mod->GetURL(), mod->GetURL(), row++, 1 );
    }

    const string& lang = mod->GetLanguage();
    string about = mod->GetAbout( lang );
    if ( about.empty() )
        about = mod->GetAbout( "en" ); // use English as fallback

    if ( !about.empty() )
    {
        makeLabel( table, _("Description:"), row, 0 );
        makeLabel( table, about, row++, 1 );
    }

    replaceTab( Notebook, ABOUT_TAB, scroll );
}

//--------------------------------------------------------------------------//
//  Update module name and usage at top of window
//--------------------------------------------------------------------------//
void
ModuleGuiBuilder::updateNameAndUsage( const Module* mod )
{
    GtkLabel* modname = GTK_LABEL( lookup_widget(TopWidget, "modname") );
    gtk_label_set_text( modname, mod->GetName().c_str() );

    const string& lang = mod->GetLanguage();
    string usageStr = mod->GetUsage( lang );
    if ( usageStr.empty() )
        usageStr = mod->GetUsage( "en" ); // use English as fallback

    GtkLabel* usage = GTK_LABEL( lookup_widget(TopWidget, "usage") );
    gtk_label_set_text( usage, usageStr.c_str() );
}

//--------------------------------------------------------------------------//
//  Dynamically fill in the notebook tabs with Module data
//--------------------------------------------------------------------------//
void
ModuleGuiBuilder::BuildGui( const Module* mod, const string& options )
{
    Notebook = GTK_NOTEBOOK( lookup_widget(TopWidget, "notebook") );
    CurrentModule = mod;

    // Forget the previous list of widgets
    WidgetMap.clear();

    // Parse the options string, and turn it into a form we can use
    StringStringMap_t optMap;
    parseOptions( options, optMap );

    if ( Verbosity )
        cout << "BuildGui: " << mod->GetName()
             << "'s options=\"" << options << "\"" << endl;

    buildParametersTab( mod, optMap );
    buildSetupTab( mod );
    buildUpdatesTab( mod );
    buildAboutTab( mod );
    updateNameAndUsage( mod );

    gtk_notebook_set_current_page( Notebook, 0 );
}

//--------------------------------------------------------------------------//
//  Parse an options string
//--------------------------------------------------------------------------//
void
ModuleGuiBuilder::parseOptions( const string& options,
                                StringStringMap_t& optMap ) const
{
    istringstream istr( options );
    istream_iterator<string> is( istr ), isend;
    for( ; is != isend; ++is )
    {
        const string& opt = *is;
        int equal = opt.find_first_of("=");
        string paramName = opt.substr( 0, equal );
        string paramVal = opt.substr( equal+1, opt.length() );
        optMap[ paramName ] = paramVal;
    }
}

//--------------------------------------------------------------------------//
//  Generate an option string for modules.conf
//--------------------------------------------------------------------------//
string
ModuleGuiBuilder::GenerateOptionsString() const
{
    ostringstream oss;
    MapParamWidget::const_iterator it = WidgetMap.begin();
    MapParamWidget::const_iterator itend = WidgetMap.end();
    for( ; it != itend; ++it )
    {
        const Param* param = (*it).first;
        const GtkWidget* widget = (*it).second;
        const string& defaultVal = param->GetDefault();

        // When a parameter is set to its default it is not written out
        switch( param->GetType() )
        {
            case PARAM_BOOL:
            {
                if ( gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget)) )
                {
                    if ( defaultVal != "1" )
                        oss << " " << param->GetName() << "=1";
                }
                else
                {
                    if ( defaultVal == "1" )
                        oss << " " << param->GetName() << "=0";
                }
                break;
            }

            case PARAM_ARRAY:
            case PARAM_INT:
            case PARAM_STRING:
            {
                const gchar* text = gtk_entry_get_text( GTK_ENTRY(widget) );
                if ( strcmp( text, "" ) && strcmp( text, defaultVal.c_str() ) )
                    oss << " " << param->GetName() << "=" << text;
                break;
            }
        }
    }

    return oss.str();
}

//--------------------------------------------------------------------------//
//  D'tor
//--------------------------------------------------------------------------//
ModuleGuiBuilder::~ModuleGuiBuilder()
{
    delete TreeViewMgr;
}

